// Copyright 2000 by Robert Dick.
// All rights reserved.

#ifndef EPSILON_H_
#define EPSILON_H_

// Epsilon computation.  For now, only works for floats and doubles.

/*###########################################################################*/
#include <limits>
#include <cmath>

#ifdef ROB_DEBUG
#	include "RStd.h"
#endif

#include "RFunctional.h"
#include "Interface.h"

namespace rstd {
/*###########################################################################*/
template <typename T>
	bool eps_is_equal_to(const T & a, const T & b, const T & ref);

template <typename T>
	bool eps_is_less(const T & a, const T & b, const T & ref);

/*===========================================================================*/
template <typename T>
struct eps_equal_to : rbinary_function<const T &, const T &, bool> {
	eps_equal_to(T ref = 0);
	bool operator()(const T & a, const T & b) const;

private:
	T ref_;
};

/*===========================================================================*/
template <typename T>
struct eps_less : rbinary_function<const T &, const T &, bool> {
	eps_less(T ref = 0);
	bool operator()(const T & a, const T & b) const;

private:
	T ref_;
};

/*===========================================================================*/
template <typename T>
struct eps_not_equal_to : rbinary_function<const T &, const T &, bool> {
	eps_not_equal_to(T ref = 0);
	bool operator()(const T & a, const T & b) const;

	T ref_;
};

template <typename T>
struct eps_greater : rbinary_function<const T &, const T &, bool> {
	eps_greater(T ref = 0);
	bool operator()(const T & a, const T & b) const;

	T ref_;
};

template <typename T>
struct eps_less_equal : rbinary_function<const T &, const T &, bool> {
	eps_less_equal(T ref = 0);
	bool operator()(const T & a, const T & b) const;

	T ref_;
};

template <typename T>
struct eps_greater_equal : rbinary_function<const T &, const T &, bool> {
	eps_greater_equal(T ref = 0);
	bool operator()(const T & a, const T & b) const;

	T ref_;
};

/*===========================================================================*/
template <typename T>
struct eps_inc : runary_function<const T &, T> {
	eps_inc(long step);
	void operator()(T & a) const;

	T step_;
};

/*===========================================================================*/
template <typename T>
struct eps_dec : runary_function<const T &, T> {
	eps_dec(long step);
	void operator()(T & a) const;

	T step_;
};

/*###########################################################################*/
template <typename T>
class eps_comp_obj {
public:
	eps_comp_obj(T ref = 0) : lt_(ref) {}
	comp_type operator()(const T & a, const T & b) const;

private:
		eps_less<T> lt_;
};

/*===========================================================================*/
template <typename T, T ref>
	comp_type eps_comp(const T & a, const T & b);

/*===========================================================================*/
void Epsilon_test();

/*###########################################################################*/
#include "Epsilon.cct"
}
#endif

